"""Implementation of GbaIpoptProblem class.

Peter Schubert, HHU Duesseldorf, June 2022
"""

import numpy as np


# based on cyipopt examples
class GbaIpoptProblem:

    def __init__(self, gba_problem):
        """Initialize GbaIpoptProblem.

        :param gba_problem: GbaProblem from where to extract relevant parameters
        :type gba_problem: GbaProblem
        """
        self.gba_problem = gba_problem
        self._n_vars = len(gba_problem.var_initial)
        self.report_freq = 0

        self.nfev = 0
        self.njev = 0
        self.nit = 0
        tmp_x = np.ones(self._n_vars) * .01
        self.constraint_dims = np.array([len(gba_problem.heq_mass_balance(tmp_x)),
                                         len(gba_problem.heq_enz_trans(tmp_x)),
                                         len(gba_problem.heq_density(tmp_x))])

    def objective(self, x):
        self.nfev += 1
        return -self.gba_problem.get_growth_rate(x)

    def gradient(self, x):
        self.njev += 1
        return -self.gba_problem.get_growth_rate_grad(x)

    def constraints(self, x):
        return np.concatenate((self.gba_problem.heq_mass_balance(x),
                               self.gba_problem.heq_enz_trans(x),
                               self.gba_problem.heq_density(x)))

    def jacobian(self, x):
        return (np.vstack((self.gba_problem.heq_mass_balance_jac(x),
                           self.gba_problem.heq_enz_trans_jac(x),
                           self.gba_problem.heq_density_jac(x)))
                ).flatten()

    def hessianstructure(self):
        return np.tril_indices(self._n_vars)

    def hessian(self, x, lagrange, obj_factor):
        hess = obj_factor * -self.gba_problem.get_growth_rate_hess(x)
        hesss = np.vstack((self.gba_problem.heq_mass_balance_hess(x),
                           self.gba_problem.heq_enz_trans_hess(x),
                           self.gba_problem.heq_density_hess(x)))
        hess += (lagrange.reshape((-1, 1, 1)) * hesss).sum(axis=0)
        return hess[np.tril_indices(self._n_vars)]

    def set_report_freq(self, mod_val=0):
        self.report_freq = mod_val

    # noinspection PyUnusedLocal
    def intermediate(self, alg_mod, iter_count, obj_value, inf_pr, inf_du, mu,
                     d_norm, regularization_size, alpha_du, alpha_pr,
                     ls_trials):
        self.nit = iter_count
        if self.report_freq > 0:
            if iter_count % self.report_freq == 0:
                print(f'[{iter_count:5d}] growth rate: {-3600 *obj_value:.5f}')
